/*************************************************************************
 Copyright (c) Shai Shalev-Shwartz.
 
 >>> SOURCE LICENSE >>>
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation (www.fsf.org); either version 2 of the
 License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 The ALGLIB project is copyrighted by Sergey Bochkanov.
 
 A copy of the GNU General Public License is available at
 http://www.fsf.org/licensing/licenses
 >>> END OF LICENSE >>>
 *************************************************************************/




//=============================================================================
// File Name: SignGECO.cc
// Written by: Shai Shalev-Shwartz (17.01.11)
// This code implements GECO for collaborating Filtering
//=============================================================================
#ifndef _SHAI_GECO_
#define _SHAI_GECO_

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <cstdlib>
#include "cmd_line.h"
#include <vector>
#include "linalg.h"
#include "solvers.h"


#define _min(x,y) ( ((x) < (y)) ? (x) : (y) )
#define _max(x,y) ( ((x) > (y)) ? (x) : (y) )

typedef unsigned int uint;

double calc_loss(uint *Itrain, uint *Jtrain, double *Valtrain, 
		 uint trainLength,alglib::real_2d_array &U, alglib::real_2d_array &V,
		 uint rr,alglib::real_1d_array &lambda,double &bias,
		 double MIN_Y,double MAX_Y);
double calc_loss_hec(uint* Itrain, uint* Jtrain, double* Valtrain, 
		     uint trainLength,alglib::real_2d_array &U, 
		     alglib::real_2d_array &V,uint rr,alglib::real_1d_array &lambda,
		     double &bias,double MIN_Y,double MAX_Y);
void sign_power_method(alglib::real_1d_array &u, uint m, alglib::real_1d_array &v, uint n,
		       uint *Itrain, uint *Jtrain, double *Grad, 
		       uint trainLength, uint NUM_POWER_ITER);
void power_method(alglib::real_1d_array &u, uint m, alglib::real_1d_array &v, uint n,
		  uint* Itrain, uint* Jtrain, double* Grad, 
		  uint trainLength, uint NUM_POWER_ITER);
void optimize(uint* Itrain, uint* Jtrain, double* Valtrain, 
	      uint trainLength,alglib::real_2d_array &U, alglib::real_2d_array &V,uint i,
	      alglib::real_2d_array &P,alglib::real_2d_array &Qt,
	      alglib::real_1d_array &lam,double &bias,double reg);
void GramSchmidtLast(alglib::real_2d_array &U, uint m,uint i);

void zeros(alglib::real_1d_array &x,uint d) {
  for (uint i=0; i<d;++i) x(i)=0.0;
}

void zeros(alglib::real_2d_array &A,uint m,uint n) {

  for (uint q=0;q<m;q++) 
    for (uint p=0;p<n;p++) 
      A(q,p) = 0.0; 
}

void prod(alglib::real_2d_array &U,uint m,uint i,alglib::real_2d_array &P,uint r,alglib::real_2d_array &sU) {
  for (uint p=0;p<m;++p) {
    for (uint q=0;q<r;++q) {
      sU(p,q) = 0.0;
      for (uint ll=0;ll<i;++ll)
	sU(p,q) += U(p,ll)*P(ll,q);
    }
  }
}

void prod2t(alglib::real_2d_array &U,uint m,uint i,alglib::real_2d_array &P,uint r,alglib::real_2d_array &sU) {
  for (uint p=0;p<m;++p) {
    for (uint q=0;q<r;++q) {
      sU(p,q) = 0.0;
      for (uint ll=0;ll<i;++ll)
	sU(p,q) += U(p,ll)*P(q,ll);
    }
  }
}


int main(int argc, char** argv) {

  // PARSE COMMAND LINE
  // ====================================================================
  // check for the correct number of command line arguments and print usage
  // message
  std::string train_filename;
  std::string test_filename;
  uint k;
  uint maxIter;
  uint NUM_POWER_ITER;
  double reg;

  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("GECO");
  cmdline.add_master_option("<train_filename>", 
			    &train_filename);
  cmdline.add_master_option("<test_filename>", 
			    &test_filename);
  cmdline.add("-k", "rank consrtaint [default 20]", &k, 20);
  cmdline.add("-b","maximal number of iterations between rank enlarge [default 10]",
	      &maxIter,20);
  cmdline.add("-p","How many power method iterations [default 30]",
	      &NUM_POWER_ITER,30);
  cmdline.add("-r","Regularization parameter for the least squares [default 1e-8]",
	      &reg,1e-8);


  int rc = cmdline.parse(argc, argv);
  if (rc < 2) {
    cmdline.print_help();
    return EXIT_FAILURE;
  }


  // OPEN DATA FILES
  // =========================
  std::ifstream train_file(train_filename.c_str());
  if (!train_file.good()) {
    std::cerr << "error w/ train_file" << std::endl;
    return EXIT_FAILURE;
  }
  std::ifstream test_file(test_filename.c_str());
  if (!test_file.good()) {
    std::cerr << "error w/ test_file" << std::endl;
    return EXIT_FAILURE;
  }

  // READ DATA
  // ===============
  uint trainLength,m,n;
  train_file >> m >> n >> trainLength;
  uint * Itrain = new uint[trainLength];
  uint * Jtrain = new uint[trainLength];
  double * Valtrain = new double[trainLength];
  for (uint i=0; i<trainLength; ++i) {
    train_file >> Itrain[i] >> Jtrain[i] >> Valtrain[i];
  }


  uint testLength;
  test_file >> m >> n >> testLength;
  uint * Itest = new uint[testLength];
  uint * Jtest = new uint[testLength];
  double * Valtest = new double[testLength];
  for (uint i=0; i<testLength; ++i) {
    test_file >> Itest[i] >> Jtest[i] >> Valtest[i];
  }


  // INITIALIZATIONS
  // ===============

  double MAX_Y = -1e6, MIN_Y = 1e6; double avgY = 0.0;
  for (uint i=0; i<trainLength; ++i) {
    if (MAX_Y < Valtrain[i]) MAX_Y = Valtrain[i]; 
    if (MIN_Y > Valtrain[i]) MIN_Y = Valtrain[i]; 
    avgY += Valtrain[i];
  }
  avgY /= trainLength;


  std::vector<double> trainLoss;
  std::vector<double> testLoss;
  std::vector<uint> AllRank; 
  uint nInter=0;

  // We'll maintain the matrix A using U,V, where A(i,j) = U(i,:)*diag(lambda)*V(j,:)'
  alglib::real_2d_array U; U.setlength(m,k+1); zeros(U,m,k+1);
  alglib::real_2d_array V; V.setlength(n,k+1); zeros(V,n,k+1);
  alglib::real_1d_array lambda; lambda.setlength(k+1); zeros(lambda,k+1);
  // We'll also maintain an additional bias term
  double bias = avgY;

  // help variables
  alglib::real_1d_array u; u.setlength(m);
  alglib::real_1d_array v; v.setlength(n);  
  alglib::real_2d_array sU; sU.setlength(m,k+1);
  alglib::real_2d_array sV; sV.setlength(n,k+1);
  alglib::real_2d_array sU2; sU2.setlength(m,k+1);
  alglib::real_2d_array sV2; sV2.setlength(n,k+1);
  alglib::real_2d_array P; 
  alglib::real_2d_array Qt;
  alglib::real_1d_array lam; 
  alglib::real_1d_array lam2; 
  double sBias,sBias2;


  // The gradient vector
  double * Grad = new double[trainLength];


  // Current rank
  uint i = 0; 

  // Main Loop
  
  for (uint iter=0; i <= k; iter++) {

    // increase rank
    i = i+1;

    // calculate gradient elements
    // For each (i,j,Y(i,j)) gradient is 2*(A(i,j)-Y(i,j)) 
    if (i>1) {
      for (uint j=0; j<trainLength; ++j) {
	double tmp = 0;
	for (uint q=0; q < i-1; ++q)
	  tmp += U(Itrain[j],q)*V(Jtrain[j],q)*lambda(q);
	Grad[j] = 2.0*( tmp + bias  - Valtrain[j]);
      }
    }
    else {
      for (uint j=0; j<trainLength; ++j) {
	Grad[j] = 2.0 * (bias - Valtrain[j]);
      }
    }
 
    // Power iteration with sign
    sign_power_method(u,m,v,n,Itrain,Jtrain,Grad,trainLength,NUM_POWER_ITER);
    for (uint q=0; q < m; ++q) U(q,i-1) = u(q);
    if (i>1) GramSchmidtLast(U,m,i); 
    sign_power_method(v,n,u,m,Jtrain,Itrain,Grad,trainLength,NUM_POWER_ITER);
    for (uint q=0; q < n; ++q) V(q,i-1) = v(q);
    if (i>1) GramSchmidtLast(V,n,i); 


    // Optimize B over the new support
    optimize(Itrain,Jtrain,Valtrain,trainLength,U,V,i,P,Qt,lam,sBias,reg);
    // That's it, now B = P*Qt

    // Update U and V
    if (i > 1 && nInter < maxIter) { // try not to increase but to replace
      prod(U,m,i,P,i-1,sU); // sU = U(1:m,1:i)*P(1:i,1:i-1)
      prod2t(V,n,i,Qt,i-1,sV); // sV = V(1:n,1:i)*Qt(1:i-1,1:i)'
      double curLoss = calc_loss_hec(Itrain,Jtrain,Valtrain,trainLength,sU,sV,i-1,lam,sBias,MIN_Y,MAX_Y);
      //double curLoss = calc_loss(Itrain,Jtrain,Valtrain,trainLength,sU,sV,i-1,sBias,MIN_Y,MAX_Y);
      i--;
      nInter++;
      if (curLoss < trainLoss[iter-1]) {
	bias = sBias;
	for (uint p=0;p<m;++p) 
	  for (uint q=0;q<i;++q) 
	    U(p,q) = sU(p,q);
	for (uint p=0;p<n;++p) 
	  for (uint q=0;q<i;++q) 
	    V(p,q) = sV(p,q);
	for (uint q=0; q<i; ++q) lambda(q) = lam(q);
	trainLoss.push_back(curLoss);
      } else {
	trainLoss.push_back(trainLoss[iter-1]);
	//nInter = maxIter;
	nInter *= 2;
      }
    }
    else { // Should increase the rank
      nInter = 0;

      // Calculate the loss if update with the sign power method
      prod(U,m,i,P,i,sU); // sU = U(1:m,1:i)*P(1:i,1:i)
      prod2t(V,n,i,Qt,i,sV); // sV = V(1:n,1:i)*Qt(1:i,1:i)'
      double sign_loss = calc_loss_hec(Itrain,Jtrain,Valtrain,trainLength,sU,sV,i,lam,sBias,MIN_Y,MAX_Y);
      //double sign_loss = calc_loss(Itrain,Jtrain,Valtrain,trainLength,sU,sV,i,sBias,MIN_Y,MAX_Y);

      // Calculate the loss if update with the regular power method
      power_method(u,m,v,n,Itrain,Jtrain,Grad,trainLength,NUM_POWER_ITER);
      for (uint q=0; q < m; ++q) U(q,i-1) = u(q);
      GramSchmidtLast(U,m,i); 
      power_method(v,n,u,m,Jtrain,Itrain,Grad,trainLength,NUM_POWER_ITER);
      for (uint q=0; q < n; ++q) V(q,i-1) = v(q);
      GramSchmidtLast(V,n,i); 
      optimize(Itrain,Jtrain,Valtrain,trainLength,U,V,i,P,Qt,lam2,sBias2,reg);
      prod(U,m,i,P,i,sU2); // sU2 = U(1:m,1:i)*P(1:i,1:i)
      prod2t(V,n,i,Qt,i,sV2); // sV2 = V(1:n,1:i)*Qt(1:i,1:i)'
      double power_loss = calc_loss_hec(Itrain,Jtrain,Valtrain,trainLength,sU2,sV2,i,lam2,sBias2,MIN_Y,MAX_Y);
      //double power_loss = calc_loss(Itrain,Jtrain,Valtrain,trainLength,sU2,sV2,i,sBias2,MIN_Y,MAX_Y);

      if (sign_loss < power_loss) {
	trainLoss.push_back(sign_loss);
	bias = sBias;
	for (uint p=0;p<m;++p) for (uint q=0;q<i;++q) U(p,q)=sU(p,q);
	for (uint p=0;p<n;++p) for (uint q=0;q<i;++q) V(p,q)=sV(p,q);
	for (uint q=0;q<i;++q) lambda(q)=lam(q);
      } else {
	trainLoss.push_back(power_loss);
	bias = sBias2;
	for (uint p=0;p<m;++p) for (uint q=0;q<i;++q) U(p,q)=sU2(p,q);
	for (uint p=0;p<n;++p) for (uint q=0;q<i;++q) V(p,q)=sV2(p,q);
 	for (uint q=0;q<i;++q) lambda(q)=lam2(q);
      }
 
    }


    testLoss.push_back(calc_loss(Itest,Jtest,Valtest,testLength,U,V,i,lambda,bias,MIN_Y,MAX_Y));
    AllRank.push_back(i);

    std::cout << "iter=" << iter << ",rank=" << AllRank[iter] << ",trainLoss=" << trainLoss[iter] << ",testLoss=" << testLoss[iter] << "\n";


    std::cout.flush(); 
    
  }

  return(EXIT_SUCCESS);

}



void rand(alglib::real_1d_array &x,uint d) {
  for (uint i=0; i<d;++i) x(i)=((double)rand()) / RAND_MAX;
}


void sign_rand(alglib::real_1d_array &x,uint d) {
  for (uint i=0; i<d;++i) x(i)=(rand() % 2)*2.0-1.0;
}

void sign(alglib::real_1d_array &x,uint d) {
  for (uint i=0; i<d;++i) x(i)=(x(i)>0)? (1) : (-1);
}

double norm(alglib::real_1d_array &x,uint d) {
  double a=0.0;
  for (uint i=0; i<d;++i) a += (x(i)*x(i)); 
  return(sqrt(a));
}

void normalize(alglib::real_1d_array &x,uint d) {
  double a=norm(x,d); 
  for (uint i=0; i<d;++i) x(i) /= a;
}


// Assuming U(:,1:i-1) are orthonormal, we set U(:,i) to be orthogonal to U(:,1:i-1)
void GramSchmidtLast(alglib::real_2d_array &U, uint m,uint i) {

  for (uint j=0; j < i-1; ++j) {
    double ip=0.0,nrm=0.0;
    for (uint p=0; p<m; ++p) {
      nrm += U(p,j)*U(p,j);
      ip += U(p,j)*U(p,i-1);
    }
    double f = ip/nrm;
    for (uint p=0; p<m; ++p) {
      U(p,i-1) -= f*U(p,j);
    }
  }
  double nrm=0.0;
  for (uint p=0; p<m; ++p) {
    nrm += U(p,i-1)*U(p,i-1);
  }
  nrm = sqrt(nrm);
  for (uint p=0; p<m; ++p) {
    U(p,i-1) /= nrm; 
  }

//   // sanity check
//   for (uint jj=0; jj < i; ++jj) {
//     for (uint j=0; j < i; ++j) {
//       double ip=0.0;
//       for (uint p=0; p<m; ++p) {
// 	ip += U(p,j)*U(p,jj);
//       }
//       std::cerr << ip << "\t\t";
//     }  
//     std::cerr << "\n";
//   }
//   std::cerr << "End GramSchmidt\n"; 

}

void sign_power_method(alglib::real_1d_array &u, uint m, alglib::real_1d_array &v, uint n,
		       uint* Itrain, uint* Jtrain, double* Grad, 
		       uint trainLength, uint NUM_POWER_ITER) {
  //  Perform power method to find the new u and v
  //  The idea, if A = sparse(Itrain,Jtrain,Grad,m,n), then:
  //  u = A*A'*...*A*A'*A*A'*randVec
  //  v = A'*A*...*A'*A*A'*A*randVec
    
  sign_rand(u,m); 

  for (uint pmIter=0; pmIter < NUM_POWER_ITER; pmIter++) {
    // set v = sign(A'*u)
    zeros(v,n);
    for (uint j=0; j <trainLength; ++j) {
      v(Jtrain[j]) += Grad[j]*u(Itrain[j]);
    }
    sign(v,n);
    // set u = sign(A*v)
    zeros(u,m);
    for (uint j=0; j <trainLength; ++j) {
      u(Itrain[j]) += Grad[j]*v(Jtrain[j]);
    }
    sign(u,m); 
  }
}

void power_method(alglib::real_1d_array &u, uint m, alglib::real_1d_array &v, uint n,
		  uint* Itrain, uint* Jtrain, double* Grad, 
		  uint trainLength, uint NUM_POWER_ITER) {
  //  Perform power method to find the new u and v
  //  The idea, if A = sparse(Itrain,Jtrain,Grad,m,n), then:
  //  u = A*A'*...*A*A'*A*A'*randVec
  //  v = A'*A*...*A'*A*A'*A*randVec
    
  rand(u,m); for (uint p=0; p<m; ++p) u(p) = (u(p)-0.5)*2.0;


  for (uint pmIter=0; pmIter < NUM_POWER_ITER; pmIter++) {
    // set v = (A'*u)
    zeros(v,n);
    for (uint j=0; j <trainLength; ++j) {
      v(Jtrain[j]) += Grad[j]*u(Itrain[j]);
    }
    // set u = A*v
    zeros(u,m);
    for (uint j=0; j <trainLength; ++j) {
      u(Itrain[j]) += Grad[j]*v(Jtrain[j]);
    }
//     if (pmIter == 0)
//       std::cerr << "norm(u) = " << norm(u,m) << "\n";
    normalize(u,m); 
//     if (pmIter == 0)
//       std::cerr << "norm(u) = " << norm(u,m) << "\n";
  }
}

double calc_loss(uint* Itrain, uint* Jtrain, double* Valtrain, 
		 uint trainLength,alglib::real_2d_array &U, alglib::real_2d_array &V,
		 uint rr, alglib::real_1d_array &lambda,double &bias,double MIN_Y,double MAX_Y) {

  double L=0;
  for (uint j=0; j<trainLength; ++j) {
    double pred = bias; for (uint p=0; p<rr; ++p) pred += U(Itrain[j],p)*V(Jtrain[j],p)*lambda(p);
    if (pred > MAX_Y) pred = MAX_Y;
    if (pred < MIN_Y) pred = MIN_Y;
    L += (pred-Valtrain[j])*(pred-Valtrain[j])/trainLength;
  }

  return(sqrt(L));
}


double calc_loss_hec(uint* Itrain, uint* Jtrain, double* Valtrain, 
		     uint trainLength,alglib::real_2d_array &U, 
		     alglib::real_2d_array &V,uint rr,alglib::real_1d_array &lambda,
		     double &bias,double MIN_Y,double MAX_Y) {

  double L=0;
  for (uint j=0; j<trainLength; ++j) {
    double pred = bias; for (uint p=0; p<rr; ++p) pred += U(Itrain[j],p)*V(Jtrain[j],p)*lambda(p);
    double cur_loss = (pred-Valtrain[j])*(pred-Valtrain[j]);
    if (pred >= MAX_Y && Valtrain[j] == MAX_Y) cur_loss = 0.0; 
    if (pred <= MIN_Y && Valtrain[j] == MIN_Y) cur_loss = 0.0; 
    L += cur_loss/trainLength;
  }

  return(sqrt(L));
}


void optimize(uint* Itrain, uint* Jtrain, double* Valtrain, 
	      uint trainLength,alglib::real_2d_array &U, alglib::real_2d_array &V,uint i,
	      alglib::real_2d_array &P,alglib::real_2d_array &Qt,
	      alglib::real_1d_array &lam,double &bias,double reg) {
	       

    // Optimize B over the new support:
    // This is a least squares problem, whose solution is 
    //  vec(B) = inv(C)*b, where 
    //  C = sum_{i,j in E} c{i,j}*c{i,j}'
    //       where c{i,j} = vec(U(i,:)'*V(j,:))
    //  b = sum_{i,j in E} Y(i,j)*c{i,j}
    uint ii= i*i+1;
    alglib::real_2d_array C; C.setlength(ii,ii); zeros(C,ii,ii); 
    alglib::real_1d_array b; b.setlength(ii); zeros(b,ii); 
    alglib::real_1d_array temp; temp.setlength(ii);

    for (uint j=0; j<trainLength; ++j) {
      for (uint p=0; p < i; ++p) {
	for (uint q=0; q < i; ++q) {
	  temp(p+q*i) = U(Itrain[j],p)*V(Jtrain[j],q); 
	}	
      }
      temp(ii-1) = 1.0; // last place for the bias term
      for (uint p=0; p < ii; ++p) {
	b(p) += Valtrain[j]*temp(p)/trainLength;
	for (uint q=0; q < ii; ++q) {
	  C(p,q) += temp(p)*temp(q)/trainLength;
	}
      }
    }

    // call solve
    alglib::ae_int_t info;
    alglib::densesolverreport rep;
    alglib::real_1d_array xx; xx.setlength(ii);
    for (uint p=0;p<ii;++p) C(p,p) += reg;
    


    rmatrixsolve(C, ii, b , info , rep, xx);

    // get the bias
    bias = xx(ii-1);


    // SVD
    alglib::real_2d_array B; B.setlength(i,i);
    for (uint p=0;p<i;++p) for (uint q=0;q<i;++q) B(p,q)=xx(p+i*q);
    //alglib::real_1d_array lam; 
    lam.setlength(i);
    P.setlength(i,i);
    Qt.setlength(i,i);
    rmatrixsvd(B, i,i, 2, 2, 2, lam, P, Qt); // now B = P*diag(lam)*Qt


//     for (uint p=0; p<i; ++p)
//       for (uint q=0; q<i; ++q)
// 	P(p,q) *= lam(q);


//     // for debug
//     std::cerr << "\nOptimize --- check orthonormality of P\n";
//     for (uint p=0; p<i; ++p) {
//       for (uint q=0; q<i; ++q) {
// 	double ip=0.0;
// 	for (uint t=0; t<i; ++t) ip+=P(p,t)*P(q,t); 
// 	std::cerr << ip << " ";
//       }
//       std::cerr << "\n";
//     }
//     std::cerr << "Done.\n\n";


}


#endif
//*****************************************************************************
//                                     E O F
//*****************************************************************************

